import inspect
import sys

from venusian.compat import walk_packages
from venusian.advice import getFrameInfo

ATTACH_ATTR = '__venusian_callbacks__'

class Scanner(object):
    def __init__(self, **kw):
        self.__dict__.update(kw)

    def scan(self, package, categories=None):
        """ Scan a Python package and any of its subpackages.  All
        top-level objects will be considered; those marked with
        venusian callback attributes related to ``category`` will be
        processed.

        The ``package`` argument should be a reference to a Python
        package or module object.

        The ``categories`` argument should be sequence of Venusian
        callback categories (each category usually a string) or the
        special value ``None`` which means all Venusian callback
        categories.  The default is ``None``.
        """
        def invoke(name, ob):
            category_keys = categories
            attached_categories = getattr(ob, ATTACH_ATTR, {})
            if category_keys is None:
                category_keys = attached_categories.keys()
                category_keys.sort()
            for category in category_keys:
                callbacks = attached_categories.get(category, [])
                for callback in callbacks:
                    callback(self, name, ob)

        for name, ob in inspect.getmembers(package):
            invoke(name, ob)

        if hasattr(package, '__path__'): # package, not module
            results = walk_packages(package.__path__, package.__name__+'.')

            for importer, modname, ispkg in results:
                __import__(modname)
                module = sys.modules[modname]
                for name, ob in inspect.getmembers(module, None):
                    invoke(name, ob)

class AttachInfo(object):
    """
    An instance of this class is returned by the
    :func:`venusian.attach` function.  It has the following
    attributes:

    ``scope``

      One of ``exec``, ``module``, ``class``, ``function call`` or
      ``unknown`` (each a string).  This is the scope detected while
      executing the decorator which runs the attach function.

    ``module``

      The module in which the decorated function was defined.

    ``locals``

      A dictionary containing decorator frame's f_locals.

    ``globals``

      A dictionary containing decorator frame's f_globals.

    ``category``

      The ``category`` argument passed to ``attach`` (or ``None``, the
      default).
      
    """
    def __init__(self, **kw):
        self.__dict__.update(kw)

def attach(wrapped, callback, category=None, depth=1):

    """ Attach a callback to the wrapped object.  It will be found
    later during a scan.  This function returns an instance of the
    :class:`venusian.AttachInfo` class."""

    frame = sys._getframe(depth+1)
    scope, module, f_locals, f_globals = getFrameInfo(frame)
    if scope == 'class':
        # we're in the midst of a class statement
        categories = f_locals.setdefault(ATTACH_ATTR, {})
        callbacks = categories.setdefault(category, [])
        callbacks.append(callback)
    else:
        categories = getattr(wrapped, ATTACH_ATTR, {})
        callbacks = categories.setdefault(category, [])
        callbacks.append(callback)
        setattr(wrapped, ATTACH_ATTR, categories)
    return AttachInfo(
        scope=scope, module=module, locals=f_locals, globals=f_globals,
        category=category)

    
    



import unittest

class TestScanner(unittest.TestCase):
    def _makeOne(self, **kw):
        from venusian import Scanner
        return Scanner(**kw)

    def test_package(self):
        from venusian.tests import fixtures
        class Test(object):
            def __init__(self):
                self.registrations = []
            def __call__(self, **kw):
                self.registrations.append(kw)
        test = Test()
        scanner = self._makeOne(test=test)
        scanner.scan(fixtures)
        self.assertEqual(len(test.registrations), 6)
        test.registrations.sort(
            lambda x, y: cmp((x['name'], x['ob'].__module__),
                             (y['name'], y['ob'].__module__))
            )
        from venusian.tests.fixtures.module import function as func1
        from venusian.tests.fixtures.module2 import function as func2
        from venusian.tests.fixtures.module import inst as inst1
        from venusian.tests.fixtures.module2 import inst as inst2
        from venusian.tests.fixtures.module import Class as Class1
        from venusian.tests.fixtures.module2 import Class as Class2

        self.assertEqual(test.registrations[0]['name'], 'Class')
        self.assertEqual(test.registrations[0]['ob'], Class1)
        self.assertEqual(test.registrations[0]['method'], True)

        self.assertEqual(test.registrations[1]['name'], 'Class')
        self.assertEqual(test.registrations[1]['ob'], Class2)
        self.assertEqual(test.registrations[1]['method'], True)

        self.assertEqual(test.registrations[2]['name'], 'function')
        self.assertEqual(test.registrations[2]['ob'], func1)
        self.assertEqual(test.registrations[2]['function'], True)

        self.assertEqual(test.registrations[3]['name'], 'function')
        self.assertEqual(test.registrations[3]['ob'], func2)
        self.assertEqual(test.registrations[3]['function'], True)
        
        self.assertEqual(test.registrations[4]['name'], 'inst')
        self.assertEqual(test.registrations[4]['ob'], inst1)
        self.assertEqual(test.registrations[4]['instance'], True)

        self.assertEqual(test.registrations[5]['name'], 'inst')
        self.assertEqual(test.registrations[5]['ob'], inst2)
        self.assertEqual(test.registrations[5]['instance'], True)

    def test_module(self):
        from venusian.tests.fixtures import module
        class Test(object):
            def __init__(self):
                self.registrations = []
            def __call__(self, **kw):
                self.registrations.append(kw)
        test = Test()
        scanner = self._makeOne(test=test)
        scanner.scan(module)
        self.assertEqual(len(test.registrations), 3)
        test.registrations.sort(
            lambda x, y: cmp((x['name'], x['ob'].__module__),
                             (y['name'], y['ob'].__module__))
            )
        from venusian.tests.fixtures.module import function as func1
        from venusian.tests.fixtures.module import inst as inst1
        from venusian.tests.fixtures.module import Class as Class1

        self.assertEqual(test.registrations[0]['name'], 'Class')
        self.assertEqual(test.registrations[0]['ob'], Class1)
        self.assertEqual(test.registrations[0]['method'], True)

        self.assertEqual(test.registrations[1]['name'], 'function')
        self.assertEqual(test.registrations[1]['ob'], func1)
        self.assertEqual(test.registrations[1]['function'], True)

        self.assertEqual(test.registrations[2]['name'], 'inst')
        self.assertEqual(test.registrations[2]['ob'], inst1)
        self.assertEqual(test.registrations[2]['instance'], True)

    def test_one_category(self):
        from venusian.tests import categoryfixture
        class Test(object):
            def __init__(self):
                self.registrations = []
            def __call__(self, **kw):
                self.registrations.append(kw)
        test = Test()
        scanner = self._makeOne(test=test)
        scanner.scan(categoryfixture, categories=('mycategory',))
        self.assertEqual(len(test.registrations), 1)
        self.assertEqual(test.registrations[0]['name'], 'function')
        self.assertEqual(test.registrations[0]['ob'], categoryfixture.function)
        self.assertEqual(test.registrations[0]['function'], True)

    def test_all_categories_implicit(self):
        from venusian.tests import categoryfixture
        class Test(object):
            def __init__(self):
                self.registrations = []
            def __call__(self, **kw):
                self.registrations.append(kw)
        test = Test()
        scanner = self._makeOne(test=test)
        scanner.scan(categoryfixture)
        self.assertEqual(len(test.registrations), 2)
        self.assertEqual(test.registrations[0]['name'], 'function')
        self.assertEqual(test.registrations[0]['ob'], categoryfixture.function)
        self.assertEqual(test.registrations[0]['function'], True)
        self.assertEqual(test.registrations[1]['name'], 'function2')
        self.assertEqual(test.registrations[1]['ob'], categoryfixture.function2)
        self.assertEqual(test.registrations[1]['function'], True)

    def test_all_categories_explicit(self):
        from venusian.tests import categoryfixture
        class Test(object):
            def __init__(self):
                self.registrations = []
            def __call__(self, **kw):
                self.registrations.append(kw)
        test = Test()
        scanner = self._makeOne(test=test)
        scanner.scan(categoryfixture, categories=('mycategory', 'mycategory2'))
        self.assertEqual(len(test.registrations), 2)
        self.assertEqual(test.registrations[0]['name'], 'function')
        self.assertEqual(test.registrations[0]['ob'], categoryfixture.function)
        self.assertEqual(test.registrations[0]['function'], True)
        self.assertEqual(test.registrations[1]['name'], 'function2')
        self.assertEqual(test.registrations[1]['ob'], categoryfixture.function2)
        self.assertEqual(test.registrations[1]['function'], True)
